# -*- coding: utf-8 -*-
from scipy.sparse import coo_matrix
import numpy as np
from numpy import ndarray

from neumann import squeeze
from neumann.array import isintegerarray, isfloatarray, \
    isboolarray, bool_to_float, atleastnd
from neumann.linalg.sparse.utils import lower_spdata, upper_spdata

from .utils import nodes2d_to_dofs1d, irows_icols_bulk, nodal_mass_matrix_data


def fem_coeff_matrix_coo(A: ndarray, *args, inds: ndarray = None,
                         rows: ndarray = None, cols: ndarray = None,
                         N: int = None, **kwargs):
    """
    Returns the coefficient matrix in sparse 'coo' format.
    Index mapping is either specified with `inds` or with
    both `rows` and `cols`.

    If `lower` or `upper` is provided as a positional argument, 
    the result is returned as a lower- or upper-triangular matrix, 
    respectively. 

    Parameters
    ----------
    A : np.ndarray[:, :, :]
        Element coefficient matrices in dense format.

    inds : np.ndarray[:, :], optional
        Global indices. The default is None.

    rows : np.ndarray[:], optional
        Global numbering of the rows. The default is None.

    cols : np.ndarray[:], optional
        Global numbering of the columns. The default is None.

    N : int, optional
        Total number of coefficients in the system. If not provided, 
        it is inferred from index data.

    Returns
    -------
    scipy.sparse.coo_matrix
        Coefficient matrix in sparse coo format (see scipy doc for the details).

    """
    if N is None:
        assert inds is not None, "shape or `inds` must be provided!."
        N = inds.max() + 1
    if rows is None:
        assert inds is not None, "Row and column indices (`rows` and `cols`) " \
            "or global index numbering array (inds) must be provided!"
        rows, cols = irows_icols_bulk(inds)

    data, rows, cols = A.flatten(), rows.flatten(), cols.flatten()

    if len(args) > 0:
        if 'lower' in args:
            data, rows, cols = lower_spdata(data, rows, cols)
        elif 'upper' in args:
            data, rows, cols = upper_spdata(data, rows, cols)

    return coo_matrix((data, (rows, cols)), shape=(N, N))


def build_fem_nodal_data(*args, inds: ndarray = None, values: ndarray = None,
                         N: int = None, **kwargs):
    """
    Returns nodal data of any sort in standard form.

    The data is given with 'values' and applies to the node indices specified
    by 'inds'. At all times, the function must be informed about the total number of 
    nodes in the model, to return an output with a correct shape. If 'inds' is None, it 
    is assumed that 'values' contains data for each node in the model, and therefore
    the length of 'values' is the number of nodes in the model. Otherwise, the number of nodes
    is inferred from 'inds' and 'N' (see the doc). 

    Parameters
    ----------

    inds : np.ndarray(int), optional
        1d numpy integer array specifying global indices for the rows of 'values'. 

    values : np.ndarray([float]), optional
        2d or 3d numpy array of floats, that specify some sort of data 
        imposed on nodes.
        If 3d, we assume that last axis stands for the number of load cases.

    N : int, optional
        The overall number of nodes in the model. If not provided, we assume that
        it equals the highest index in 'inds' + 1 (this assumes zero-based indexing).

    Notes
    -----
    The function returns the values as a 2d array, even if the input array was only 
    2 dimensional, which suggests a single load case. 

    Call 'fem_load_vector' with the same arguments to get the vector itself as a numpy array. 

    Returns
    -------
    (indices, nodal_loads, N) or None

        indices : int(:)
            1d numpy integer array of npde indices

        nodal_loads : float(:, :) 
            2d numpy float array of shape (nN * nDOF, nRHS)

        N : int
            The overall size of the equation system, as derived from the input.

    """

    assert isfloatarray(values)

    # node based definition with multiple RHS (at least 1)
    values = atleastnd(values, 3, back=True)  # (nP, nDOF, nRHS)
    if inds is None:
        inds = np.arange(len(values))  # node indices
    else:
        assert isintegerarray(inds)
        assert len(values) == len(inds)
    # transform to nodal defintion
    inds, values = nodes2d_to_dofs1d(inds, values)

    N = inds.max() + 1 if N is None else N
    return inds, values, N


@squeeze(True)
def fem_load_vector(*args, **kwargs) -> ndarray:
    """
    Assembles the right-hand-side of the global equation system.

    Parameters
    ----------
        See the documentation of 'build_fem_nodal_data' for the discription
        of the possible arguments.

    Returns
    -------
    numpy.ndarray(float)[:]
        The load vector as a numpy array of floats.
    """
    inds, values, N = build_fem_nodal_data(*args, **kwargs)
    nRHS = values.shape[-1]
    f = np.zeros((N, nRHS))
    f[inds] = values
    return f


def build_fem_ebc_data(*args, inds: ndarray = None, values: ndarray = None,
                       pfix: float = 1e12, **kwargs):
    """
    Returns fixity information in standard form. It extends the behaviour of 
    'build_fem_nodal_data' by allowing for boolean input.

    Parameters
    ----------
    inds : see 'build_fem_nodal_data'

    values : see 'build_fem_nodal_data'

    pfix : float, optional
        Penalty value for fixed dofs. It is used to transform boolean penalty
        data, or to make up for missing values (e.g. only indices are provided).
        Default value is 1e+12.

    Notes
    -----
    It is used to create a penalty matrix for a Courant-type penalization of the 
    essential boundary conditions.

    Call 'fem_penalty_matrix_coo' with the same arguments to get the penalty 
    matrix itself as a sparse scipy matrix in 'coo' format. 

    Returns
    -------
    see 'build_fem_nodal_data'

    """
    if values is not None:
        assert isinstance(values, np.ndarray)
        values = atleastnd(values, 3, back=True)  # (nP, nDOF, nLHS)
    else:
        assert isintegerarray(inds)
        raise NotImplementedError

    if isfloatarray(values):
        pass
    elif isboolarray(values):
        values = bool_to_float(values, pfix)
    else:
        raise NotImplementedError

    return build_fem_nodal_data(*args, inds=inds, values=values, **kwargs)


def fem_penalty_matrix_coo(*args, eliminate_zeros=True, sum_duplicates=True,
                           **kwargs) -> coo_matrix:
    """
    Returns the sparse, COO format penalty matrix, equivalent of 
    a Courant-type penalization of the essential boundary conditions.

    Parameters
    ----------
        See the documentation of 'build_fem_ebc_data' for the discription
        of the possible arguments.

    Returns
    -------
    scipy.sparse.coo_matrix
        The penalty matrix in sparse COO format.
    """
    inds, pen, N = build_fem_ebc_data(*args, **kwargs)
    K = coo_matrix((pen[:, 0], (inds, inds)), shape=(N, N))
    if eliminate_zeros:
        K.eliminate_zeros()
    if sum_duplicates:
        K.sum_duplicates()
    return K


def fem_nodal_mass_matrix_coo(*args, values=None, eliminate_zeros=True, 
                              sum_duplicates=True, ndof=6, **kwargs) -> coo_matrix:
    """
    Returns the diagonal mass matrix resulting form nodal masses in scipy COO format.

    Parameters
    ----------
        See the documentation of 'build_fem_ebc_data' for the discription
        of the possible arguments.

    Returns
    -------
    scipy.sparse.coo_matrix
        The mass matrix in sparse COO format.
    
    """
    values = nodal_mass_matrix_data(values, ndof)
    inds, vals, N = build_fem_nodal_data(*args, values=values, **kwargs)
    M = coo_matrix((vals[:, 0], (inds, inds)), shape=(N, N))
    if eliminate_zeros:
        M.eliminate_zeros()
    if sum_duplicates:
        M.sum_duplicates()
    return M


if __name__ == '__main__':
    inds = np.array([2, 4, 12])
    pen = np.array([1e5, 1e5, 1e12])
    N = 100
    args = build_fem_ebc_data(inds=inds, values=pen, N=N)



# -*- coding: utf-8 -*-

from ..config import __hassympy__, __haspylatex__

if __hassympy__ and __haspylatex__:
    import sympy as sy
    from sympy import symbols, Function, diff, Matrix, MatMul, \
        integrate, Symbol, sin, cos, pi, simplify
    import pylatex as pltx
    from pylatex import Section, Subsection, Command, NoEscape, NewPage

from .latex import expr_to_ltx, expr_to_ltx_breqn, append_preamble
from .symtools import inv_sym_3x3


def append_mindlin_theory(doc, *args, glossary:dict=None, signs:dict=None, **kwargs):
    
    if glossary is None:
        glossary = {
            'UX' : 'u_0',
            'UY' : 'v_0',
            'UZ' : 'w_0',
            'ROTX' : '\\theta_y',
            'ROTY' : '\\theta_x',
            'ROTZ' : '\\theta_z',
        }
        
    if signs is None:
        glossary = {
            'UX' : 1,
            'UY' : 1,
            'UZ' : 1,
            'ROTX' : 1,
            'ROTY' : 1,
            'ROTZ' : 1,
        }

    __hasreqs__ = __hassympy__ and __haspylatex__
    assert __hasreqs__, "You must have `sympy` and `pylatex` for this."

    with doc.create(Section('Theoretical background')):

        # coordinates in a suitable vector space
        x, y, z = symbols('x y z', real=True)

        # kinematic variables
        UZ = Function(glossary['UZ'])(x, y)
        ROTX = Function(glossary['ROTY'])(x, y)
        ROTY = Function(glossary['ROTX'])(x, y)

        # sign modifiers for kinematic variables
        sign_w, sign_u, sign_v, = symbols('\\beta_w, \\beta_u, \\beta_v', real=True)

        # displacement field
        with doc.create(Subsection('Displacement Field and Boundary Conditions')):
            doc.append(NoEscape(
                r"""
                \noindent
                The Mindlin-Reissner kinematical model poses the following
                assumptions on the displacement field:
                """))
            doc.append(NoEscape(
                r"""
                \begin{enumerate}[label*=\protect\fbox{MR\arabic{enumi}},
                itemindent=5em]
                \item   normals to the midsurface remain normal
                \item   the plate thickness does not change during deformation
                \item   normal stress through the thickness is ignored
                \end{enumerate}
                """))

            """
            u = -z * beta_u * thx
            v = -z * beta_v * thy
            w = beta_w * w0
            """
            u = -z * ROTX
            v = -z * ROTY
            w = UZ
            doc.append(expr_to_ltx(r'u(x, y, z)', sy.latex(u)))
            doc.append(expr_to_ltx(r'v(x, y, z)', sy.latex(v)))
            doc.append(expr_to_ltx(r'w(x, y, z)', sy.latex(w)))

        # engineering strains
        with doc.create(Subsection('Strain - Displacement Equations')):
            exx = diff(u, x)
            eyy = diff(v, y)
            exy = diff(u, y) + diff(v, x)
            exz = diff(w, x) + diff(u, z)
            eyz = diff(w, y) + diff(v, z)
            doc.append(expr_to_ltx(r'\varepsilon_x(x, y, z)', sy.latex(exx)))
            doc.append(expr_to_ltx(r'\varepsilon_y(x, y, z)', sy.latex(eyy)))
            doc.append(expr_to_ltx(r'\gamma_{xy}(x, y, z)', sy.latex(exy)))
            doc.append(expr_to_ltx(r'\gamma_{xz}(x, y, z)', sy.latex(exz)))
            doc.append(expr_to_ltx(r'\gamma_{yz}(x, y, z)', sy.latex(eyz)))

            doc.append('Curvatures are defined as')
            kxx = diff(exx, z)
            kyy = diff(eyy, z)
            kxy = diff(exy, z)
            doc.append(
                expr_to_ltx(r"""\kappa_x(x, y) \coloneqq \frac{\partial
                \varepsilon_x}{\partial z}""", sy.latex(kxx)))
            doc.append(
                expr_to_ltx(r"""\kappa_y(x, y) \coloneqq \frac{\partial
                \varepsilon_y}{\partial z}""", sy.latex(kyy)))
            doc.append(
                expr_to_ltx(r"""\kappa_{xy}(x, y) \coloneqq \frac{\partial
                \gamma_{xy}}{\partial z}""", sy.latex(kxy)))

        # material model and stress resultants
        with doc.create(Subsection('Stress Resultants')):
            # strains '+' Hooke model -> stresses
            C11, C22, C12, C16, C26, C66 = C126ij = \
                symbols('C_11 C_22 C_12 C_16 C_26 C_66', real=True)
            C126 = sy.Matrix([[C11, C12, 0], [C12, C22, 0], [0, 0, C66]])
            sxx, syy, sxy = MatMul(C126, sy.Matrix([exx, eyy, exy])).doit()
            C44, C55 = C45ij = symbols('C_44 C_55', real=True)
            C45 = sy.Matrix([[C44, 0], [0, C55]])
            syz, sxz = MatMul(C45, sy.Matrix([eyz, exz])).doit()

            # integrate through the thickness
            h = Symbol('h', real=True)  # thickness
            def Int(expr): return integrate(expr, (z, -h/2, h/2))
            mx, my, mxy = M = Matrix([Int(s * z) for s in [sxx, syy, sxy]])
            vx, vy = V = Matrix([Int(s) for s in [sxz, syz]])

            D11, D22, D12, D66 = Dij = symbols(
                'D_11 D_22 D_12 D_66', real=True)
            S44, S55, = Sij = symbols('S_44 S_55', real=True)
            #
            mx = mx.simplify().expand()
            cD11 = mx.coeff(C11 * h**3 / 12)
            cD12 = mx.coeff(C12 * h**3 / 12)
            mx = D11 * cD11 + D12 * cD12
            #
            my = my.simplify().expand()
            cD22 = my.coeff(C22 * h**3 / 12)
            cD21 = my.coeff(C12 * h**3 / 12)
            my = D22 * cD22 + D12 * cD21
            #
            mxy = mxy.simplify().expand()
            cD66 = mxy.coeff(C66 * h**3 / 12)
            mxy = D66 * cD66
            #
            vx = vx.simplify().expand()
            cS55 = vx.coeff(C55 * h)
            vx = S55 * cS55
            #
            vy = vy.simplify().expand()
            cS44 = vy.coeff(C44 * h)
            vy = S44 * cS44

            # sign modifiers for internal forces
            """
            inds = ['x', 'y', 'xy', 'xz', 'yz']
            syms = " ".join(["\\alpha_{}".format('{' + i + '}') for i in inds])
            alpha = symbols(sims, real = True)
            mx *= alpha[0]
            my *= alpha[1]
            mxy *= alpha[2]
            vx *= alpha[3]
            vy *= alpha[4]
            """

            mx = mx.simplify().expand()
            my = my.simplify().expand()
            mxy = mxy.simplify().expand()
            vx = vx.simplify().expand()
            vy = vy.simplify().expand()
            doc.append(expr_to_ltx(r'm_x(x, y)', sy.latex(mx)))
            doc.append(expr_to_ltx(r'm_y(x, y)', sy.latex(my)))
            doc.append(expr_to_ltx(r'm_{xy}(x, y)', sy.latex(mxy)))
            doc.append(expr_to_ltx(r'v_x(x, y)', sy.latex(vx)))
            doc.append(expr_to_ltx(r'v_y(x, y)', sy.latex(vy)))

        # Equilibrium equations. The signs of the equations
        # is selected such, that the coefficients of the load functions on the
        # right-hand sides admit positive signs according to global axes.
        ###
        # moment around global X
        lhs_mx = simplify(-diff(mxy, x) - diff(my, y) + vy).expand()
        # moment around global Y
        lhs_my = simplify(diff(mxy, y) + diff(mx, x) - vx).expand()
        # vertical equilibrium
        lhs_fz = simplify(diff(vx, x) + diff(vy, y)).expand()
        with doc.create(Subsection('Equilibrium Equations')):
            doc.append(NoEscape(r'Equilibrium of moments around global $x$ :'))
            doc.append(expr_to_ltx_breqn(sy.latex(lhs_mx), 'm_x(x, y)'))
            doc.append(NoEscape(r'Equilibrium of moments around global $y$ :'))
            doc.append(expr_to_ltx_breqn(sy.latex(lhs_my), 'm_y(x, y)'))
            doc.append(NoEscape(r'Equilibrium of forces along global $z$ :'))
            doc.append(expr_to_ltx_breqn(sy.latex(lhs_fz), 'f_z(x, y)'))

        # Boundary Conditions and Trial Solution
        with doc.create(Subsection('Boundary Conditions and Trial Solution')):
            mn = m, n = symbols('m n', integer=True, positive=True)
            coeffs = Amn, Bmn, Cmn = symbols('A_{mn} B_{mn} C_{mn}', real=True)
            shape = Lx, Ly = symbols('L_x, L_y', real=True)
            Sm, Sn = sin(m * pi * x / Lx), sin(n * pi * y / Ly)
            Cm, Cn = cos(m * pi * x / Lx), cos(n * pi * y / Ly)
            w0_trial = Cmn * Sm * Sn
            thx_trial = Amn * Cm * Sn
            thy_trial = Bmn * Sm * Cn
            doc.append(NoEscape(
                r"""
                All of the boundary conditions are of the essential type, and
                are satisfied a priori by the proper selection of approximation
                functions. Thus, the unknown field functions are sought in the
                form
                """))
            sumltx = r'\sum_{m, n} '
            doc.append(expr_to_ltx(sy.latex(UZ), sumltx + sy.latex(w0_trial)))
            doc.append(expr_to_ltx(sy.latex(ROTX),
                       sumltx + sy.latex(thx_trial)))
            doc.append(expr_to_ltx(sy.latex(ROTY),
                       sumltx + sy.latex(thy_trial)))

            # substitute trial solution
            trial = {UZ: w0_trial, ROTX: thx_trial, ROTY: thy_trial}
            eq_mx_trial = lhs_mx.subs(trial).expand().doit() / (Sm * Cn)
            eq_mx_trial = eq_mx_trial.simplify().expand()
            eq_my_trial = lhs_my.subs(trial).expand().doit() / (Cm * Sn)
            eq_my_trial = eq_my_trial.simplify().expand()
            eq_fz_trial = lhs_fz.subs(trial).expand().doit() / (Sm * Sn)
            eq_fz_trial = eq_fz_trial.simplify().expand()
            doc.append(NoEscape(
                r"""
                \noindent
                Upon substitution of the trial solution into the vertical
                equilibrium equation we obtain the following equality
                """))
            mxltx = r"\left[" + sy.latex(eq_fz_trial.simplify()) + r"\right]" + \
                sy.latex(Sm * Sn)
            doc.append(expr_to_ltx_breqn(mxltx, 'f_z(x, y)'))
            doc.append(NoEscape(
                r"""
                \noindent
                It is clear, that if was possible to write the function
                of vertical forces $f_z(x, y)$ as a finite sum of purely sinusoidal
                functions, the trigonometric terms could be completely ruled out
                from the equation. After forming similar thoughts about the other
                two equations, the loads are sought as
                """))

            # coefficient matrix
            doc.append('Discrete equilibrium equation')
            P = sy.zeros(3, 3)
            P[0, :] = Matrix([eq_mx_trial.coeff(c) for c in coeffs]).T
            P[1, :] = Matrix([eq_my_trial.coeff(c) for c in coeffs]).T
            P[2, :] = Matrix([eq_fz_trial.coeff(c) for c in coeffs]).T
            doc.append(expr_to_ltx(r'\boldsymbol{P}', sy.latex(P), dfrac=True))
            detP, adjP = inv_sym_3x3(P, as_adj_det=True)
            detP = detP.simplify().expand()
            adjP.simplify()

        with doc.create(Subsection('Postprocessing')):
            mx = mx.subs(trial).expand().doit()
            mx = mx.simplify().expand()
            my = my.subs(trial).expand().doit()
            my = my.simplify().expand()
            mxy = mxy.subs(trial).expand().doit()
            mxy = mxy.simplify().expand()
            vx = vx.subs(trial).expand().doit()
            vx = vx.simplify().expand()
            vy = vy.subs(trial).expand().doit()
            vy = vy.simplify().expand()
            kxx = kxx.subs(trial).expand().doit()
            kxx = kxx.simplify().expand()
            kyy = kyy.subs(trial).expand().doit()
            kyy = kyy.simplify().expand()
            kxy = kxy.subs(trial).expand().doit()
            kxy = kxy.simplify().expand()

            exz = exz.subs(trial).expand().doit()
            exz = exz.simplify().expand()
            eyz = eyz.subs(trial).expand().doit()
            eyz = eyz.simplify().expand()

            doc.append(expr_to_ltx(r'm_x(x, y)', sy.latex(mx)))
            doc.append(expr_to_ltx(r'm_y(x, y)', sy.latex(my)))
            doc.append(expr_to_ltx(r'm_{xy}(x, y)', sy.latex(mxy)))
            doc.append(expr_to_ltx(r'v_x(x, y)', sy.latex(vx)))
            doc.append(expr_to_ltx(r'v_y(x, y)', sy.latex(vy)))
            doc.append(expr_to_ltx(r'\kappa_x(x, y)', sy.latex(kxx)))
            doc.append(expr_to_ltx(r'\kappa_y(x, y)', sy.latex(kyy)))
            doc.append(expr_to_ltx(r'\kappa_{xy}(x, y)', sy.latex(kxy)))
            doc.append(expr_to_ltx(r'\gamma_{xz}(x, y)', sy.latex(exz)))
            doc.append(expr_to_ltx(r'\gamma_{yz}(x, y)', sy.latex(eyz)))

        with doc.create(Subsection('Loads')):
            doc.append(NoEscape(
                r"""
                If the load functions are represented as suggested by the
                left hand-sides of the equilibrium equations, the solution of
                the system simplifies to the solution of a 3x3 algebraic equation
                system.
                """))

            # constans vertical load over a rectangular area
            q, w, h = symbols('q w h', real=True)
            xc, yc = symbols('x_c y_c', real=True)
            qmn = (4/(Lx*Ly)) * integrate(q * Sm * Sn, (x, xc - w/2, xc + w/2),
                                          (y, yc - h/2, yc + h/2))
            qmn = qmn.simplify().expand()
            doc.append(NoEscape(
                r"""
                \noindent
                \paragraph{Constans vertical load of intensity $q$ over a
                rectangular area}
                """))
            doc.append(expr_to_ltx(r'q_{mn}', sy.latex(qmn), post=','))
            # doc.append(NoEscape(r"\noindent"))
            doc.append(NoEscape(
                r"""
                where $w$ and $h$ denote the width and height, $x_c$ and $y_c$ the
                coordinates of the center point of the rectangle.
                """))

            # constans moment of intensity mx, around global axis X,
            # over a rectangular area with width w, height h and center (xc, yc)
            m_x, w, h = symbols('m_x w h', real=True)
            xc, yc = symbols('x_c y_c', real=True)
            qmn = (4/(Lx*Ly)) * integrate(m_x * Sm * Cn, (x, xc - w/2, xc + w/2),
                                          (y, yc - h/2, yc + h/2))
            qmn = qmn.simplify().expand()
            doc.append(NoEscape(
                r"""
                \noindent
                \paragraph{Constans moment around global axis $x$ of intensity
                $m_x$ over a rectangular area}
                """))
            doc.append(expr_to_ltx(r'q_{mn}', sy.latex(qmn), post=','))
            # doc.append(NoEscape(r"\noindent"))
            doc.append(NoEscape(
                r"""
                where $w$ and $h$ denote the width and height, $x_c$ and $y_c$ the
                coordinates of the center of the rectangle.
                """))

            # constant moment of intensity my, around global axis Y,
            # over a rectangular area with width w, height h and center (xc, yc)
            m_y, w, h = symbols('m_y w h', real=True)
            xc, yc = symbols('x_c y_c', real=True)
            qmn = (4/(Lx*Ly)) * integrate(m_y * Cm * Sn, (x, xc - w/2, xc + w/2),
                                          (y, yc - h/2, yc + h/2))
            qmn = qmn.simplify().expand()
            doc.append(NoEscape(
                r"""
                \noindent
                \paragraph{Constans moment around global axis $y$ of intensity
                $m_y$ over a rectangular area}
                """))
            doc.append(expr_to_ltx(r'q_{mn}', sy.latex(qmn), post=','))
            # doc.append(NoEscape(r"\noindent"))
            doc.append(NoEscape(
                r"""
                where $w$ and $h$ denote the width and height, $x_c$ and $y_c$ the
                coordinates of the center of the rectangle.
                """))
            
            return doc

    
if __name__ == '__main__':
    assert __hassympy__ and __haspylatex__
    from sigmaepsilon.solid.fourier.latex import  append_preamble

    geometry_options = {
        "tmargin": "1.5cm",
        "lmargin": "1.5cm",
        "rmargin": "1.5cm"
    }
    
    doc = pltx.Document(geometry_options=geometry_options)

    doc = append_preamble(doc)

    title = "Calculation of rectangular, simply supported plates \
        resting on elastic foundation."
    doc.preamble.append(Command('title', title))
    doc.preamble.append(Command('author', 'Claude Louis Marie Henri Navier'))
    doc.preamble.append(Command('date', NoEscape(r'\today')))
    doc.append(NoEscape(r'\maketitle'))
    doc.append(NoEscape(r'\tableofcontents'))
    doc.append(NewPage())
    
    doc = append_mindlin_theory(doc)
    
    doc.generate_pdf('f:\\navier2', clean_tex=False, compiler='pdfLaTeX')



# -*- coding: utf-8 -*-
import numpy as np
from numpy import ndarray
from scipy.sparse import spmatrix
from scipy.sparse import coo_matrix as npcoo, csc_matrix as npcsc
import time
from typing import Union

from linkeddeepdict import LinkedDeepDict

from neumann.array import atleast2d
from .utils import irows_icols_bulk
from .linsolve import (box_fem_data_bulk, solve_standard_form, 
                       unbox_lhs)
from .dyn import (effective_modal_masses, Rayleigh_quotient, 
                  natural_circular_frequencies)

"""
PARDISO MATRIX TYPES

1
real and structurally symmetric

2
real and symmetric positive definite

-2
real and symmetric indefinite

3
complex and structurally symmetric

4
complex and Hermitian positive definite

-4
complex and Hermitian indefinite

6
complex and symmetric

11
real and nonsymmetric

13
complex and nonsymmetric

"""

class Solver:
    """ 
    Base Class for Solvers. This is only a placeholder at the moment 
    for future work to be done.
    """


class FemSolver(Solver):
    """
    A class to perform solutions for linear elastostatics
    and dynamics.

    Parameters
    ----------
    K : Union[:class:`scipy.sparse.spmatrix`, :class:`numpy.ndarray`]
        The stiffness matrix.
    
    Kp : :class:`scipy.sparse.spmatrix`
        The penalty stiffness matrix for the Courant-type penalization
        of the essential boundary conditions.
    
    f : :class:`numpy.ndarray`
        The load vector as an 1d or 2d numpy array.
        
    gnum : :class:`numpy.ndarray`
        Global dof numbering as a numpy array.
    
    imap : Union[dict, :class:`numpy.ndarray`], Optional
        Index mapper. Default is None.
        
    regular : bool, Optional
        If True it is assumed that the structure is regular. Default is True.
        
    M : Union[:class:`scipy.sparse.spmatrix`, :class:`numpy.ndarray`], Optional
        The mass matrix. Default is None.
        
    """

    def __init__(self, K : Union[spmatrix, ndarray], Kp : spmatrix, f : ndarray, 
                 gnum: ndarray, imap : Union[dict, ndarray]=None, regular:bool=True, 
                 M:Union[spmatrix, ndarray]=None, **config):
        self.K = K
        self.M = M
        self.Kp = Kp
        self.gnum = gnum
        self.f = f
        self.vmodes = None  # vibration modes : Tuple(vals, vecs)
        self.config = config
        self.regular = regular
        self.imap = imap
        if imap is not None:
            # Ff imap.shape[0] > f.shape[0] it means that the inverse of
            # the mapping is given. It would require to store more data, but
            # it would enable to tell the total size of the equation system, that the
            # input is a representation of.
            assert imap.shape[0] == f.shape[0]
        self.regular = False if imap is not None else regular
        self.core = self.encode()
        self.READY = False
        self.summary = []
        self._summary = LinkedDeepDict()
    
    def elastic_stiffness_matrix(self, sparse=True, penalize=True) -> Union[ndarray, spmatrix]:
        """
        Returns the stiffness matrix.
        
        Parameters
        ----------
        sparse : bool, Optional
            Set this true to get the result as a sparse 2d matrix, otherwise
            a 3d dense matrix is retured with matrices for all the individual elements.
            Default is True.
            
        penalize: bool, Optional
            If this is True, what is returendd is the sum of the elastic stiffness matrix
            plus the Courant-type penalty stiffness matrix of the essential boundary conditions.
            Default is True.
        
        Returns
        -------
        Union[:class:`numpy.ndarray`, :class:`scipy.linalg.spmatrix`]
            The stiffness matrix.
            
        """
        if not sparse:
            assert not penalize, "The penalty matrix is only available in sparse format."
            return self.K
        else:
            if penalize:
                return self.Ke + self.Kp
            else:
                return self.Ke
            
    def consistent_mass_matrix(self) -> spmatrix:
        """
        Returns the consistent mass matrix of the solver.
        
        Note
        ----
        The 'consistent' attribute means that when formulating
        the element matrices, we use the same shape functions that we
        use to formulate the stiffness matrix.
        
        Returns
        -------
        :class:`scipy.linalg.spmatrix`
            The mass matrix.
            
        """
        return self.M

    def encode(self) -> 'FemSolver':
        if self.imap is None and not self.regular:
            Kp, gnum, f, imap = box_fem_data_bulk(self.Kp, self.gnum, self.f)
            self.imap = imap
            return FemSolver(self.K, Kp, f, gnum, regular=True)
        else:
            return self
        
    def decode(self):
        assert not self.regular
        N = self.gnum.max() + 1
        self.u = unbox_lhs(self.core.u, self.imap, N=N)
        self.r = unbox_lhs(self.core.r, self.imap, N=N)

    def preproc(self, force=False):
        _t0 = time.time()
        if self.READY and not force:
            return
        self.N = self.gnum.max() + 1
        if self.config.get('sparsify', False):
            if not isinstance(self.f, npcsc):
                self.f = npcsc(self.f)
        self.krows, self.kcols = irows_icols_bulk(self.gnum)
        self.krows = self.krows.flatten()
        self.kcols = self.kcols.flatten()

        if not self.regular:
            self.core.preproc()
            self.Ke = self.core.Ke
        else:
            self.Ke = npcoo((self.K.flatten(), (self.krows, self.kcols)),
                            shape=(self.N, self.N), dtype=self.K.dtype)
        self.READY = True
        _dt = time.time() - _t0
        self._summary['preproc', 'regular'] = self.regular
        self._summary['preproc', 'time'] = _dt
        self._summary['preproc', 'N'] = self.N

    def update_stiffness(self, K):
        self.K = K
        self.preproc(force=True)

    def proc(self, preproc=False, solver=None):
        if preproc:
            self.preproc(force=True)
        _t0 = time.time()
        if not self.regular:
            self.core.proc()
            _dt = time.time() - _t0
            self._summary['proc']['time'] = _dt
            return self
        Kcoo = self.Ke + self.Kp
        Kcoo.eliminate_zeros()
        Kcoo.sum_duplicates()
        self.u, summary = solve_standard_form(
            Kcoo, self.f, summary=True, solver=solver)
        self._summary['proc'] = summary

    def postproc(self):
        _t0 = time.time()
        if not self.regular:
            self.core.postproc()
            self._summary['core'] = self.core._summary
            return self.decode()
        self.r = np.reshape(self.Ke.dot(self.u), self.f.shape) - self.f
        _dt = time.time() - _t0
        self._summary['postproc', 'time'] = _dt

    def linsolve(self, *args, solver=None, summary=False, **kwargs):
        """
        Performs a linear elastostatic analysis and returns the unknown
        coefficients as a numpy array.
        
        Parameters
        ----------
        solver : str, Optional.
        
        Returns
        -------
        :class:`numpy.ndarray`
            The vector of nodal displacements as a numpy array.
        
        dict, Optional
            A summary about the solution, only if 'summary' is True.
            
        """
        self._summary = LinkedDeepDict()

        self.preproc()
        self.proc(solver=solver)
        self.postproc()

        self.summary.append(self._summary)
        
        if summary:
            return self.u, self.summary
        else:
            return self.u
        
    def natural_circular_frequencies(self, *args, return_vectors=False, **kwargs):
        """
        Returns natural circular frequencies. The call forwards all parameters
        to :func:`sigmaepslion.solid.fem.dyn.natural_circular_frequencies`, 
        see the docs there for the details. 
        
        Parameters
        ----------
        return_vectors : bool, Optional
            To return eigenvectors or not. Default is False.
                
        See also
        --------
        :func:`sigmaepslion.solid.fem.dyn.natural_circular_frequencies`
        
        Returns
        -------
        :class:`numpy.ndarray`
            The natural circular frequencies.

        :class:`numpy.ndarray`
            The eigenvectors, only if 'return_vectors' is True.
            
        """
        K = self.Ke + self.Kp
        M = self.M
        self.vmodes = natural_circular_frequencies(K, M, *args, return_vectors=True, **kwargs)
        if return_vectors:
            return self.vmodes
        else:
            return self.vmodes[0] 

    def effective_modal_masses(self, *args, action=None, modes=None, **kwargs):
        """
        Returns effective modal masses of several modes of vibration. 
        The call forwards all parameters to 
        :func:`sigmaepslion.solid.fem.dyn.effective_modal_masses`, see the 
        docs there for the details.
        
        See also
        --------
        :func:`sigmaepslion.solid.fem.dyn.effective_modal_masses`
        
        Returns
        -------
        :class:`numpy.ndarray`
            An array of effective mass values.
            
        """
        if action is None:
            raise TypeError("No action is provided!")
        vecs = self.vmodes[1] if modes is None else modes
        return effective_modal_masses(self.M, action, vecs)
    
    def Rayleigh_quotient(self, *args, u=None, f=None, **kwargs) -> ndarray:
        """
        Returns Rayleigh's quotient. The call forwards all parameters
        to :func:`sigmaepslion.solid.fem.dyn.Rayleigh_quotient`, see the 
        docs there for the details. If there are no actions specified, 
        the function feeds it with the results from a linear elastic solution.
        
        See also
        --------
        :func:`sigmaepslion.solid.fem.dyn.Rayleigh_quotient`
        
        Returns
        -------
        float or :class:`numpy.ndarray`
            One or more floats.
            
        """
        M = self.M
        if isinstance(f, ndarray) and u is None:
            K = self.Ke + self.Kp
            K.eliminate_zeros()
            K.sum_duplicates()
            u = atleast2d(solve_standard_form(K, f), back=True)
        elif isinstance(u, ndarray) and f is None:
            K = self.Ke + self.Kp
            K.eliminate_zeros()
            K.sum_duplicates()
            f = K @ u
        elif f is None and u is None:
            assert self.u is not None, "A linear solution must be calculated prior to this."
            u = self.u
            return self.Rayleigh_quotient(u=u)    
        return Rayleigh_quotient(M, u, f=f)
                


# -*- coding: utf-8 -*-
import numpy as np
from numpy import swapaxes as swap
from time import time

from .loads import LoadGroup
from .preproc import lhs_Navier
from .postproc import postproc
from .proc import linsolve


class RectangularPlate:

    def __init__(self, size: tuple, shape: tuple, *args,
                 D: np.ndarray = None, S: np.ndarray = None,
                 Winkler=None, Pasternak=None, loads=None,
                 Hetenyi=None, model='mindlin', mesh=None,
                 Bernoulli: bool = None, Navier=True, **kwargs):
        assert Navier is True, "Only Navier boundary conditions are covered at the moment."
        self.size = np.array(size, dtype=float)
        self.shape = np.array(shape, dtype=int)
        self.D = D
        self.S = S
        self.model = model.lower()
        self.mesh = mesh
        if isinstance(loads, LoadGroup):
            self.loads = loads
        elif isinstance(loads, dict):
            self.add_loads_from_dict(loads)
        elif isinstance(loads, np.ndarray):
            raise NotImplementedError
        else:
            self.loads = LoadGroup(Navier=self)
        self._summary = {}
        self.Bernoulli = not self.model in [
            'm', 'mindlin'] if Bernoulli is None else Bernoulli
        assert isinstance(
            self.Bernoulli, bool), "The keyword argument `Bernoulli` must be `True` or `False`."

    def add_loads_from_dict(self, d: dict, *args, **kwargs):
        self.loads = LoadGroup.from_dict(d)
        self.loads._Navier = self
        return self.loads

    def solve(self, *args, key='_coeffs', postproc=False, **kwargs):
        self._summary = {}
        self._key_coeff = key
        LC = list(self.loads.load_cases())
        D = self.D
        if self.Bernoulli and self.model in ['m', 'mindlin']:
            D = np.full(self.D.shape, 1e12)
        LHS = lhs_Navier(self.size, self.shape, D=D, S=self.S,
                         model=self.model, squeeze=False)
        RHS = list(lc.rhs() for lc in LC)
        [setattr(lc, '_rhs', rhs) for lc, rhs in zip(LC, RHS)]
        if self.model in ['k', 'kirchhoff']:
            RHS = [np.sum(rhs, axis=-1) for rhs in RHS]
        t0 = time()
        coeffs = linsolve(LHS, np.stack(RHS), squeeze=False)
        self._summary['dt'] = time() - t0
        nRHS, nLHS, nMN = coeffs.shape[:3]
        self._summary['nMN'] = nMN
        self._summary['nRHS'] = nRHS
        self._summary['nLHS'] = nLHS
        [setattr(lc, self._key_coeff, c) for lc, c in zip(LC, coeffs)]

    def postproc(self, points: np.ndarray, *args, cleanup=True,
                 key='res2d', **kwargs):
        self._key_res2d = key
        size = self.size
        shape = self.shape
        LC = list(self.loads.load_cases())
        RHS = np.stack(list(lc._rhs for lc in LC))
        coeffs = np.stack([getattr(lc, self._key_coeff) for lc in LC])
        ABDS = np.zeros((5, 5))
        ABDS[:3, :3] = self.D
        if self.S is not None:
            ABDS[3:, 3:] = self.S
        else:
            ABDS[3:, 3:] = 1e12
        res = postproc(ABDS, points, size=size, shape=shape, loads=RHS,
                       solution=coeffs, model=self.model)
        if len(LC) == 1:
            setattr(LC[0], self._key_res2d, np.squeeze(swap(res, 0, 1)))
        else:
            [setattr(lc, self._key_res2d, np.squeeze(swap(r2d, 0, 1)))
                for lc, r2d in zip(LC, res)]
        res = np.squeeze(res)
        if cleanup:
            [delattr(lc, self._key_coeff) for lc in LC]
